﻿###############################################################################
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
###############################################################################


<#

.Synopsis
Adds or extends a custom property named ComputerUser of type string[] to an instance of a Windows Powershell object.

.Description
The Add-ComputerUser cmdlet lets you add a member named ComputerUser of type string[] to an instance of a Windows Powershell object. The source of this data depends on the parameters specified.

For example, you can use the -FromComputerManagedBy switch to look up the user for a computer in Active Directory by querying the managedBy property of the associated computer object.

To use Add-ComputerUser, pipe the output of Read-ADRecoveryInformation or Read-ADTpmInformation to Add-ComputerUser or use the InputObject parameter to specify the RecoveryInformation or TpmInformation object.

Multiple user sources may be specified simultaneously.

This file (Add-ComputerUser.ps1) is a self-describing cmdlet. You can invoke this Cmdlet in one of two ways.

1. Via the file name:
PS > Read-ADRecoveryInformation -Server contoso.com -Credential $cred -Recurse | .\Add-ComputerUser.ps1 -Server contoso.com -FromComputerManagedBy

2. Via dot sourcing the Add-ComputerUser.ps1 file to define the Add-ComputerUser function
PS > . .\Add-ComputerUser.ps1
PS > Read-ADRecoveryInformation -Server contoso.com -Credential $cred -Recurse | Add-ComputerUser -Server contoso.com -FromComputerManagedBy

.Parameter Server
The Active Directory server to query for computer user information. Only used if -FromComputerManagedBy or -FromComputerSDDLOwner switches are specified. By default, the local domain's LDAP service is queried.

.Parameter InputObject
Specifies the object to which the ComputerUser field is added or extended.

.Parameter FromComputerSDDLOwner
Gathers user information from the owner of the computer object in Active Directory as listed in the security descriptor. The computer object is located by using the value of the ComputerDistinguishedName field on the InputObject.

.Parameter FromComputerManagedBy
Gathers user information from the computer object's managedBy field in Active Directory. The computer object is located by using the value of the ComputerDistinguishedName field on the InputObject.

.Parameter FromComputerUserMapping
Specifies the computer-to-user lookup table to use as an array of PsObject objects. Each PsObject must have a property named Computer and a property named User. The InputObject must have a property named Computer.

The following example demonstrates how the computer users are obtained from the provided table.

$FromComputerUserMapping | Where-Object {$_.Computer -eq $InputObject.Computer} | Select-Object -Property User -Unique

.Example
[pscustomobject]@{Computer="maverick-pc.contoso.com"} | Add-ComputerUser -FromComputerUserMapping (ConvertFrom-Csv ('Computer,User','maverick-pc.contoso.com,maverick@contoso.com','maverick-pc.contoso.com,viper@contoso.com','hollywood-pc.contoso.com,hollywood@contoso.com'))

Populates the ComputerUser array with the accounts associated with the computer in the provided table. (i.e. 'select User from table where Computer = $_.Computer' where $_ represents the pipeline input PsObject).

This command requires the input object to have a property named Computer.

This command requires the table value of ComputerUserMapping to have a column named Computer and a column named User.

.Example
Read-ADRecoveryInformation -Server contoso.com -Credential $cred -Recurse | Add-ComputerUser -Server contoso.com -FromComputerManagedBy

Populates the ComputerUser array with the account referenced by the Computer Object's managedBy field in Active Directory.

The computer object in Active Directory is discovered by its DistinguishedName. This command requires the input object to have a property named ComputerDistinguishedName.

Use this command if it is standard operating procedure within your organization to populate this field of the computer object in Active Directory.

.Example
Read-ADRecoveryInformation -Server contoso.com -Credential $cred -Recurse | Add-ComputerUser -Server contoso.com -FromComputerSDDLOwner

Populates the ComputerUser array with the account listed as the owner of the Active Directory computer object in the object's security descriptor.

The computer object in Active Directory is discovered by its DistinguishedName. This command requires the input object to have a property named ComputerDistinguishedName.

Use this command if it is standard operating procedure within your organization for end users to self-join their company issued machines to the domain.

.Example
Read-ADRecoveryInformation -Server contoso.com -Credential $cred -Recurse | Add-ComputerUser -Server contoso.com -FromComputerSDDLOwner -FromComputerManagedBy -FromComputerUserMapping (ConvertFrom-Csv ('Computer,User','maverick-pc.contoso.com,maverick@contoso.com'))

This example combines examples 1, 2 and 3

.Example
Read-ADTpmInformation -Server contoso.com -Credential $cred -Recurse | Add-ComputerUser -FromComputerUserMapping (Import-Csv ComputerToUserMapping.csv) | Write-MBAMTpmInformation -RecoveryServiceEndPoint https://mbam-iis.contoso.com/MBAMRecoveryAndHardwareService/CoreService.svc

This is an end to end example that reads TPM OwnerAuth data from Active Directory, adds users to each TpmInformation object on the pipeline by querying the imported csv file (contains two columns: Computer and User) then pipes the result to Write-MBAMTpmInformation to import the data into the MBAM recovery database.

.Inputs
PsObject output from Read-ADRecoveryInformation or Read-ADTpmInformation

.Outputs
The InputObject with any additional users added as a ComputerUser property of type String[]

#>
[CmdletBinding()]
param (
    [parameter(Mandatory=$false)]
    [string]$Server,

    [parameter(Mandatory=$true, ValueFromPipeline=$true)]
    [PsObject]$InputObject,

    [parameter()]
    [switch]$FromComputerSDDLOwner,

    [parameter()]
    [switch]$FromComputerManagedBy,

    [parameter()]
    [PsObject[]]$FromComputerUserMapping
)

BEGIN {
Set-StrictMode -Version 1

$here = Split-Path -Parent $MyInvocation.MyCommand.Path
if (!(Test-Path "$here\ConvertFrom-DirectoryServiceObjectName.ps1")) {
    throw [System.InvalidOperationException]"Add-ComputerUser.ps1 depends on ConvertFrom-DirectoryServiceObjectName.ps1 in the same directory"
}
# dot source the definition into this script scope now. Use dummy values for mandatory parameters
. "$here\ConvertFrom-DirectoryServiceObjectName.ps1" -Name dotsource -Type SID

# Constructs a hashtable that can be used for splatting when invoking a cmdlet using only the specified paramters.
# Splatting is preferrable to invoke-expression because it helps avoid the risk of powershell injection.
Function New-ParamHashtable {
    Param(
        [string]$Identity,
        [string]$Server,
        [string[]]$Properties
    )
    $argumentList = @{};
    if (!($Identity -eq $null) -and ![string]::IsNullOrWhiteSpace($Identity)) {
        $argumentList['Identity'] = $Identity
    }
    if (!($Server -eq $null) -and ![string]::IsNullOrWhiteSpace($Server)) {
        $argumentList['Server'] = $Server
    }
    if (!($Properties -eq $null) -and $Properties.Count -gt 0) {
        $argumentList['Properties'] = $Properties
    }
    return $argumentList
} #New-ParamHashtable

# Returns the fully qualified UPN of the user identified as the owner of the computer object in Active Directory identified by $InputObject.'ComputerDistinguishedName'
# If unsuccessful, $null is returned with an error written to the error stream or one of the following exepctions are thrown (Win32Exception, InvalidOperationException)
Function Get-ADComputerSDDLOwner {
    Param(
        [parameter(Mandatory=$true)]
        [PsObject]$InputObject,

        [parameter(Mandatory=$false)]
        [string]$Server
    )
    if (![string]::IsNullOrWhiteSpace($InputObject.'ComputerDistinguishedName')) {
        [string]$path = ("LDAP://" + $InputObject.'ComputerDistinguishedName')
        if (![string]::IsNullOrWhiteSpace($Server)) {
            $path = ("LDAP://" + $Server + "/" + $InputObject.'ComputerDistinguishedName')
        }
        [string[]]$arguments = @($path)
        
        $de = New-Object System.DirectoryServices.DirectoryEntry -ArgumentList $arguments
        try {
            $de.AuthenticationType = [System.DirectoryServices.AuthenticationTypes]::Secure -bor [System.DirectoryServices.AuthenticationTypes]::Signing -bor [System.DirectoryServices.AuthenticationTypes]::Sealing;
            $objectSecurity = $de.ObjectSecurity
            [string]$ownerSid = [string]::Empty
            if (!($objectSecurity -eq $null)) {
                [string]$ownerSid = $objectSecurity.GetOwner([System.Security.Principal.SecurityIdentifier]).Value
            }
        } finally {
            $de.Dispose()
        }
        if (![string]::IsNullOrWhiteSpace($ownerSid)) {
            $ownerSid = $ownerSid.TrimStart("O:")
        }
        if (![string]::IsNullOrWhiteSpace($ownerSid)) {
            return ConvertFrom-DirectoryServiceObjectName -Name $ownerSid -Type SID
        } else {
            Write-Error -Message "Could not retrieve the owner from the security descriptor of $path." -Category ObjectNotFound -TargetObject $InputObject -ErrorId "FromComputerSDDLOwnerWithoutSDDLOwner"
        }
    } else {
        Write-Error -Message "FromComputerSDDLOwner requires pipeline input with 'ComputerDistinguishedName' property." -Category InvalidArgument -TargetObject $InputObject -ErrorId "FromComputerSDDLOwnerWithoutComputerDistinguisedName" -RecommendedAction "Pipe the output of Read-ADRecoveryInformation or Read-ADTpmInformation into this CmdLet."
    }
} #Get-ADComputerSDDLOwner

# Returns the fully qualified UPN of the user identified in the managedBy field of the computer object in Active Directory identified by $InputObject.'ComputerDistinguishedName'
# If unsuccessful, $null is returned with an error written to the error stream or one of the following exepctions are thrown (Win32Exception, InvalidOperationException)
Function Get-ADComputerManagedBy {
    Param(
        [parameter(Mandatory=$true)]
        [PsObject]$InputObject,

        [parameter(Mandatory=$false)]
        [string]$Server
    )
    if (![string]::IsNullOrWhiteSpace($InputObject.'ComputerDistinguishedName')) {
        $argumentList = New-ParamHashtable -Identity $InputObject.'ComputerDistinguishedName' -Server $Server -Properties managedBy
        $computer =  Get-ADComputer @argumentList
        if (![string]::IsNullOrWhiteSpace($computer.'managedBy')) {
            return ConvertFrom-DirectoryServiceObjectName -Name $computer.'managedBy' -Type DN
        } else {
            Write-Error -Message ("The managedBy field of the ADComputer object was not specified for the computer " + $InputObject.'ComputerDistinguishedName' + ".") -Category ObjectNotFound -TargetObject $InputObject -ErrorId "FromComputerManagedByWithoutManagedBy" -RecommendedAction "Populate the managedBy field in ActiveDirectory on the computer objects in order to use this feature."
        }
    }
    else {
        Write-Error -Message "FromComputerManagedBy requires pipeline input with 'ComputerDistinguishedName' property." -Category InvalidArgument -TargetObject $InputObject -ErrorId "FromComputerManagedByWithoutComputerDistinguisedName" -RecommendedAction "Pipe the output of Read-ADRecoveryInformation or Read-ADTpmInformation into this CmdLet."
    }
} #Get-ADComputerManagedBy

# Returns $FromComputerUserMapping | where { $_.Computer -ieq $InputObject.'Computer' } | % { $_.'User' }
# If the PsObject objects in the supplied FromComputerUserMapping array is not in the correct format, $null is returned with an error written to the error stream.
Function Get-AssociatedUsersFromMapping {
    Param(
        [parameter(Mandatory=$true)]
        [PsObject]$InputObject,

        [parameter(Mandatory=$true)]
        [PsObject[]]$FromComputerUserMapping
    )
    if ([string]::IsNullOrWhiteSpace($InputObject.'Computer')) {
        Write-Error -Message "FromComputerUserMapping requires pipeline input with 'Computer' property." -Category InvalidArgument -TargetObject $InputObject -ErrorId "FromComputerUserMappingWithoutComputer" -RecommendedAction "Pipe the output of Read-ADRecoveryInformation or Read-ADTpmInformation into this CmdLet."
        return
    }
    if ($FromComputerUserMapping.Count -gt 0) {
        if ([string]::IsNullOrWhiteSpace($FromComputerUserMapping[0].Computer)) {
            Write-Error -Message "FromComputerUserMapping requires mapping table with a column named 'Computer'." -Category InvalidArgument -TargetObject $FromComputerUserMapping -ErrorId "FromComputerUserMappingWithoutComputerColumn" -RecommendedAction "Import a CSV file with two columns, one named Computer, the other User that provides the Computer to Users mapping."
            return
        }
        if ([string]::IsNullOrWhiteSpace($FromComputerUserMapping[0].User)) {
            Write-Error -Message "FromComputerUserMapping requires mapping table with a column named 'User'." -Category InvalidArgument -TargetObject $FromComputerUserMapping -ErrorId "FromComputerUserMappingWithoutUserColumn" -RecommendedAction "Import a CSV file with two columns, one named Computer, the other User that provides the Computer to Users mapping."
            return
        }
    }
    $FromComputerUserMapping | where {
        $_.Computer -ieq $InputObject.'Computer'
    } | % {
        $_.'User'
    }
} #Get-AssociatedUsersFromMapping

# Named definition of the Add-ComputerUser cmdlet defined by this ps1 file.
Function Add-ComputerUser {
    [CmdletBinding()]
    param (
        [parameter(Mandatory=$false)]
        [string]$Server,

        [parameter(Mandatory=$true, ValueFromPipeline=$true)]
        [PsObject]$InputObject,

        [parameter()]
        [switch]$FromComputerSDDLOwner,

        [parameter()]
        [switch]$FromComputerManagedBy,

        [parameter()]
        [PsObject[]]$FromComputerUserMapping
    )
    BEGIN {}
    PROCESS {
        #Use a hashtable to deduplicate discovered users of the computer
        [hashtable]$computerUsers = @{}
        # If the InputObject already has a ComputerUser property, augment it.
        if (!($InputObject.ComputerUser -eq $null)) {
            if ($InputObject.ComputerUser.GetType() -eq [string[]]) {
                $InputObject.ComputerUser | %{ $computerUsers[$_] = $true }
            } else {
                Write-Error -Message "InputObject already has a ComputerUser property but not of type string[]" -Category InvalidData -TargetObject $InputObject
            }
        }

        if ($FromComputerSDDLOwner) {
            try {
                $username = Get-ADComputerSDDLOwner -InputObject $InputObject -Server $Server
                if (![string]::IsNullOrWhiteSpace($username)) {
                    $computerUsers[$username] = $true
                }
            } catch [InvalidOperationException] {
                Write-Error -Exception $_.Exception -Category ObjectNotFound -TargetObject $InputObject
            } catch [System.ComponentModel.Win32Exception] {
                Write-Error -Exception $_.Exception -Category ResourceUnavailable -TargetObject $InputObject
            }
        }
        if ($FromComputerManagedBy) {
            try {
                $username = Get-ADComputerManagedBy -InputObject $InputObject -Server $Server
                if (![string]::IsNullOrWhiteSpace($username)) {
                    $computerUsers[$username] = $true
                }
            } catch [InvalidOperationException] {
                Write-Error -Exception $_.Exception -Category ObjectNotFound -TargetObject $InputObject
            } catch [System.ComponentModel.Win32Exception] {
                Write-Error -Exception $_.Exception -Category ResourceUnavailable -TargetObject $InputObject
            }
        }
        if ($FromComputerUserMapping.Count -gt 0) {
            Get-AssociatedUsersFromMapping -InputObject $InputObject -FromComputerUserMapping $FromComputerUserMapping | % {
                $computerUsers[$_] = $true
            }
        }
        if ($computerUsers.Count -gt 0) {
            Add-Member -InputObject $InputObject -MemberType NoteProperty -Name "ComputerUser" -Value ([string[]]$computerUsers.Keys)
        }
        Write-Output $InputObject
    } #PROCESS

    END {}
} #Add-ComputerUser
} #BEGIN
PROCESS {
    Add-ComputerUser @PSBoundParameters
}
END {
}

# SIG # Begin signature block
# MIIauQYJKoZIhvcNAQcCoIIaqjCCGqYCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUjmHiZs/10AzXLcSg6Vh3ecMg
# ZOagghWCMIIEwzCCA6ugAwIBAgITMwAAAG9lLVhtBxFGKAAAAAAAbzANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTUwMzIwMTczMjAy
# WhcNMTYwNjIwMTczMjAyWjCBszELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
# bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
# b3JhdGlvbjENMAsGA1UECxMETU9QUjEnMCUGA1UECxMebkNpcGhlciBEU0UgRVNO
# OkMwRjQtMzA4Ni1ERUY4MSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBT
# ZXJ2aWNlMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAz+ZtzcEqza6o
# XtiVTy0DQ0dzO7hC0tBXmt32UzZ31YhFJGrIq9Bm6YvFqg+e8oNGtirJ2DbG9KD/
# EW9m8F4UGbKxZ/jxXpSGqo4lr/g1E/2CL8c4XlPAdhzF03k7sGPrT5OaBfCiF3Hc
# xgyW0wAFLkxtWLN/tCwkcHuWaSxsingJbUmZjjo+ZpWPT394G2B7V8lR9EttUcM0
# t/g6CtYR38M6pR6gONzrrar4Q8SDmo2XNAM0BBrvrVQ2pNQaLP3DbvB45ynxuUTA
# cbQvxBCLDPc2Ynn9B1d96gV8TJ9OMD8nUDhmBrtdqD7FkNvfPHZWrZUgNFNy7WlZ
# bvBUH0DVOQIDAQABo4IBCTCCAQUwHQYDVR0OBBYEFPKmSSl4fFdwUmLP7ay3eyA0
# R9z9MB8GA1UdIwQYMBaAFCM0+NlSRnAK7UD7dvuzK7DDNbMPMFQGA1UdHwRNMEsw
# SaBHoEWGQ2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3Rz
# L01pY3Jvc29mdFRpbWVTdGFtcFBDQS5jcmwwWAYIKwYBBQUHAQEETDBKMEgGCCsG
# AQUFBzAChjxodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jv
# c29mdFRpbWVTdGFtcFBDQS5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZI
# hvcNAQEFBQADggEBAI2zTLbY7A2Hhhle5ADnl7jVz0wKPL33VdP08KCvVXKcI5e5
# girHFgrFJxNZ0NowK4hCulID5l7JJWgnJ41kp235t5pqqz6sQtAeJCbMVK/2kIFr
# Hq1Dnxt7EFdqMjYxokRoAZhaKxK0iTH2TAyuFTy3JCRdu/98U0yExA3NRnd+Kcqf
# skZigrQ0x/USaVytec0x7ulHjvj8U/PkApBRa876neOFv1mAWRDVZ6NMpvLkoLTY
# wTqhakimiM5w9qmc3vNTkz1wcQD/vut8/P8IYw9LUVmrFRmQdB7/u72qNZs9nvMQ
# FNV69h/W4nXzknQNrRbZEs+hm63SEuoAOyMVDM8wggTsMIID1KADAgECAhMzAAAB
# Cix5rtd5e6asAAEAAAEKMA0GCSqGSIb3DQEBBQUAMHkxCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNp
# Z25pbmcgUENBMB4XDTE1MDYwNDE3NDI0NVoXDTE2MDkwNDE3NDI0NVowgYMxCzAJ
# BgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25k
# MR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xDTALBgNVBAsTBE1PUFIx
# HjAcBgNVBAMTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJL8bza74QO5KNZG0aJhuqVG+2MWPi75R9LH7O3HmbEm
# UXW92swPBhQRpGwZnsBfTVSJ5E1Q2I3NoWGldxOaHKftDXT3p1Z56Cj3U9KxemPg
# 9ZSXt+zZR/hsPfMliLO8CsUEp458hUh2HGFGqhnEemKLwcI1qvtYb8VjC5NJMIEb
# e99/fE+0R21feByvtveWE1LvudFNOeVz3khOPBSqlw05zItR4VzRO/COZ+owYKlN
# Wp1DvdsjusAP10sQnZxN8FGihKrknKc91qPvChhIqPqxTqWYDku/8BTzAMiwSNZb
# /jjXiREtBbpDAk8iAJYlrX01boRoqyAYOCj+HKIQsaUCAwEAAaOCAWAwggFcMBMG
# A1UdJQQMMAoGCCsGAQUFBwMDMB0GA1UdDgQWBBSJ/gox6ibN5m3HkZG5lIyiGGE3
# NDBRBgNVHREESjBIpEYwRDENMAsGA1UECxMETU9QUjEzMDEGA1UEBRMqMzE1OTUr
# MDQwNzkzNTAtMTZmYS00YzYwLWI2YmYtOWQyYjFjZDA1OTg0MB8GA1UdIwQYMBaA
# FMsR6MrStBZYAck3LjMWFrlMmgofMFYGA1UdHwRPME0wS6BJoEeGRWh0dHA6Ly9j
# cmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY0NvZFNpZ1BDQV8w
# OC0zMS0yMDEwLmNybDBaBggrBgEFBQcBAQROMEwwSgYIKwYBBQUHMAKGPmh0dHA6
# Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljQ29kU2lnUENBXzA4LTMx
# LTIwMTAuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCmqFOR3zsB/mFdBlrrZvAM2PfZ
# hNMAUQ4Q0aTRFyjnjDM4K9hDxgOLdeszkvSp4mf9AtulHU5DRV0bSePgTxbwfo/w
# iBHKgq2k+6apX/WXYMh7xL98m2ntH4LB8c2OeEti9dcNHNdTEtaWUu81vRmOoECT
# oQqlLRacwkZ0COvb9NilSTZUEhFVA7N7FvtH/vto/MBFXOI/Enkzou+Cxd5AGQfu
# FcUKm1kFQanQl56BngNb/ErjGi4FrFBHL4z6edgeIPgF+ylrGBT6cgS3C6eaZOwR
# XU9FSY0pGi370LYJU180lOAWxLnqczXoV+/h6xbDGMcGszvPYYTitkSJlKOGMIIF
# vDCCA6SgAwIBAgIKYTMmGgAAAAAAMTANBgkqhkiG9w0BAQUFADBfMRMwEQYKCZIm
# iZPyLGQBGRYDY29tMRkwFwYKCZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQD
# EyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRob3JpdHkwHhcNMTAwODMx
# MjIxOTMyWhcNMjAwODMxMjIyOTMyWjB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALJyWVwZMGS/HZpgICBC
# mXZTbD4b1m/My/Hqa/6XFhDg3zp0gxq3L6Ay7P/ewkJOI9VyANs1VwqJyq4gSfTw
# aKxNS42lvXlLcZtHB9r9Jd+ddYjPqnNEf9eB2/O98jakyVxF3K+tPeAoaJcap6Vy
# c1bxF5Tk/TWUcqDWdl8ed0WDhTgW0HNbBbpnUo2lsmkv2hkL/pJ0KeJ2L1TdFDBZ
# +NKNYv3LyV9GMVC5JxPkQDDPcikQKCLHN049oDI9kM2hOAaFXE5WgigqBTK3S9dP
# Y+fSLWLxRT3nrAgA9kahntFbjCZT6HqqSvJGzzc8OJ60d1ylF56NyxGPVjzBrAlf
# A9MCAwEAAaOCAV4wggFaMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFMsR6MrS
# tBZYAck3LjMWFrlMmgofMAsGA1UdDwQEAwIBhjASBgkrBgEEAYI3FQEEBQIDAQAB
# MCMGCSsGAQQBgjcVAgQWBBT90TFO0yaKleGYYDuoMW+mPLzYLTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTAfBgNVHSMEGDAWgBQOrIJgQFYnl+UlE/wq4QpTlVnk
# pDBQBgNVHR8ESTBHMEWgQ6BBhj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtp
# L2NybC9wcm9kdWN0cy9taWNyb3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEE
# SDBGMEQGCCsGAQUFBzAChjhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2Nl
# cnRzL01pY3Jvc29mdFJvb3RDZXJ0LmNydDANBgkqhkiG9w0BAQUFAAOCAgEAWTk+
# fyZGr+tvQLEytWrrDi9uqEn361917Uw7LddDrQv+y+ktMaMjzHxQmIAhXaw9L0y6
# oqhWnONwu7i0+Hm1SXL3PupBf8rhDBdpy6WcIC36C1DEVs0t40rSvHDnqA2iA6VW
# 4LiKS1fylUKc8fPv7uOGHzQ8uFaa8FMjhSqkghyT4pQHHfLiTviMocroE6WRTsgb
# 0o9ylSpxbZsa+BzwU9ZnzCL/XB3Nooy9J7J5Y1ZEolHN+emjWFbdmwJFRC9f9Nqu
# 1IIybvyklRPk62nnqaIsvsgrEA5ljpnb9aL6EiYJZTiU8XofSrvR4Vbo0HiWGFzJ
# NRZf3ZMdSY4tvq00RBzuEBUaAF3dNVshzpjHCe6FDoxPbQ4TTj18KUicctHzbMrB
# 7HCjV5JXfZSNoBtIA1r3z6NnCnSlNu0tLxfI5nI3EvRvsTxngvlSso0zFmUeDord
# EN5k9G/ORtTTF+l5xAS00/ss3x+KnqwK+xMnQK3k+eGpf0a7B2BHZWBATrBC7E7t
# s3Z52Ao0CW0cgDEf4g5U3eWh++VHEK1kmP9QFi58vwUheuKVQSdpw5OPlcmN2Jsh
# rg1cnPCiroZogwxqLbt2awAdlq3yFnv2FoMkuYjPaqhHMS+a3ONxPdcAfmJH0c6I
# ybgY+g5yjcGjPa8CQGr/aZuW4hCoELQ3UAjWwz0wggYHMIID76ADAgECAgphFmg0
# AAAAAAAcMA0GCSqGSIb3DQEBBQUAMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAX
# BgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290
# IENlcnRpZmljYXRlIEF1dGhvcml0eTAeFw0wNzA0MDMxMjUzMDlaFw0yMTA0MDMx
# MzAzMDlaMHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAf
# BgNVBAMTGE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQTCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJ+hbLHf20iSKnxrLhnhveLjxZlRI1Ctzt0YTiQP7tGn
# 0UytdDAgEesH1VSVFUmUG0KSrphcMCbaAGvoe73siQcP9w4EmPCJzB/LMySHnfL0
# Zxws/HvniB3q506jocEjU8qN+kXPCdBer9CwQgSi+aZsk2fXKNxGU7CG0OUoRi4n
# rIZPVVIM5AMs+2qQkDBuh/NZMJ36ftaXs+ghl3740hPzCLdTbVK0RZCfSABKR2YR
# JylmqJfk0waBSqL5hKcRRxQJgp+E7VV4/gGaHVAIhQAQMEbtt94jRrvELVSfrx54
# QTF3zJvfO4OToWECtR0Nsfz3m7IBziJLVP/5BcPCIAsCAwEAAaOCAaswggGnMA8G
# A1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFCM0+NlSRnAK7UD7dvuzK7DDNbMPMAsG
# A1UdDwQEAwIBhjAQBgkrBgEEAYI3FQEEAwIBADCBmAYDVR0jBIGQMIGNgBQOrIJg
# QFYnl+UlE/wq4QpTlVnkpKFjpGEwXzETMBEGCgmSJomT8ixkARkWA2NvbTEZMBcG
# CgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5ghB5rRahSqClrUxzWPQHEy5lMFAGA1UdHwRJ
# MEcwRaBDoEGGP2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1
# Y3RzL21pY3Jvc29mdHJvb3RjZXJ0LmNybDBUBggrBgEFBQcBAQRIMEYwRAYIKwYB
# BQUHMAKGOGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljcm9z
# b2Z0Um9vdENlcnQuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEB
# BQUAA4ICAQAQl4rDXANENt3ptK132855UU0BsS50cVttDBOrzr57j7gu1BKijG1i
# uFcCy04gE1CZ3XpA4le7r1iaHOEdAYasu3jyi9DsOwHu4r6PCgXIjUji8FMV3U+r
# kuTnjWrVgMHmlPIGL4UD6ZEqJCJw+/b85HiZLg33B+JwvBhOnY5rCnKVuKE5nGct
# xVEO6mJcPxaYiyA/4gcaMvnMMUp2MT0rcgvI6nA9/4UKE9/CCmGO8Ne4F+tOi3/F
# NSteo7/rvH0LQnvUU3Ih7jDKu3hlXFsBFwoUDtLaFJj1PLlmWLMtL+f5hYbMUVbo
# nXCUbKw5TNT2eb+qGHpiKe+imyk0BncaYsk9Hm0fgvALxyy7z0Oz5fnsfbXjpKh0
# NbhOxXEjEiZ2CzxSjHFaRkMUvLOzsE1nyJ9C/4B5IYCeFTBm6EISXhrIniIh0EPp
# K+m79EjMLNTYMoBMJipIJF9a6lbvpt6Znco6b72BJ3QGEe52Ib+bgsEnVLaxaj2J
# oXZhtG6hE6a/qkfwEm/9ijJssv7fUciMI8lmvZ0dhxJkAj0tr1mPuOQh5bWwymO0
# eFQF1EEuUKyUsKV4q7OglnUa2ZKHE3UiLzKoCG6gW4wlv6DvhMoh1useT8ma7kng
# 9wFlb4kLfchpyOZu6qeXzjEp/w7FW1zYTRuh2Povnj8uVRZryROj/TGCBKEwggSd
# AgEBMIGQMHkxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xIzAh
# BgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBAhMzAAABCix5rtd5e6as
# AAEAAAEKMAkGBSsOAwIaBQCggbowGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQw
# HAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwIwYJKoZIhvcNAQkEMRYEFAjS
# /TqM2ytXxbS2kusIwdAgv5DsMFoGCisGAQQBgjcCAQwxTDBKoCyAKgBNAGkAYwBy
# AG8AcwBvAGYAdAAgAEMAbwByAHAAbwByAGEAdABpAG8AbqEagBhodHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEBBQAEggEAERruxu+Ndqw3kK5K1Jgh
# OOmZWTseT2fIqo07e3C7d5kDCtRLB8/R2qGcBVGn9fkZySTZ++C+FMM4q9A3BmL/
# Ti6QKdNVSkpkXokqO483G5/0LgICioeWJYeAlB/QL/i3Zw3M9Qra+kDy7U5V71yy
# JjOuAUYwdXPNsMzdZbaMd22cTsf46trVg3xJsH/7bnJYkwm9ooe4SkYEDiqFxPPK
# 9eM1O6h7TIDawXfAk0CyvUy1pt6XB5dCdKskooBqqDx5reK93/56rIZBvoCL9QP0
# +7Gv8Bzonsm/s//UIZI5iCPfXIVHirbVwAlcdbhwc5F/G6GycQX2f+Dml6+9UpYE
# wKGCAigwggIkBgkqhkiG9w0BCQYxggIVMIICEQIBATCBjjB3MQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhNaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBQQ0ECEzMAAABvZS1YbQcRRigAAAAAAG8wCQYFKw4DAhoFAKBdMBgG
# CSqGSIb3DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTE1MDcxNjE5
# Mzg0NlowIwYJKoZIhvcNAQkEMRYEFHXp2NbCPp87R9cDZBHoJfPCDpZeMA0GCSqG
# SIb3DQEBBQUABIIBAIVI1rxV3SgFXjYXK+IAVQMD4uqOSFaZz7Uu0C8nQBUBK1ck
# lJnfbFmTVWa9KJgcmywjia9qvIYRn0sWfN+trNrjDyFPp0tvsepkNg3PLp6z3Woo
# 1m09JYRoNR6nXTQcs/R5t7XgKG3MeM8kiC0imrxve9VTA3uBAHd60I+2nVQWalt0
# +ovlkbnoclKq2Nyz7JGz/bvFM7VenL+KWsNYvhkJ0cnNrnYhbJsAzb9+9U0moxmL
# 9nABMselF+HofwC/IW74VgaUE7OgGU5dKrwUAJf8iOC9Gz6ZxAYt2OeHe3Y53MQd
# FdKd6LfgS//I8XaFFbeEyD3/Seb0OAorRLgXyu8=
# SIG # End signature block
