﻿###############################################################################
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
###############################################################################


<#

.Synopsis
Converts a SID or a DistinguishedName of a directory service object into a fully qualified user principal name.

.Description
The ConvertFrom-DirectoryServiceObjectName cmdlet allows you to get a fully qualified UPN (i.e. user@subdomain.company.com) from a SID or from a Active Directory distinguished name.
If the ConvertFrom-DirectoryServiceObjectName cmdlet fails, no value is returned and either a Win32Exception or an InvalidOperationException is thrown.

This file (ConvertFrom-DirectoryServiceObjectName.ps1) is a self-describing cmdlet. You can invoke this Cmdlet in one of two ways.

1. Via the file name:
PS > .\ConvertFrom-DirectoryServiceObjectName.ps1 -Name 'S-1-5-21-397955417-626881126-188441444-501' -Format SID

2. Via dot sourcing the ConvertFrom-DirectoryServiceObjectName.ps1 file to define the ConvertFrom-DirectoryServiceObjectName function
PS > . .\ConvertFrom-DirectoryServiceObjectName.ps1 -Name 'dotsource' -Format SID
PS > ConvertFrom-DirectoryServiceObjectName -Name 'CN=user,OU=UserAccounts,DC=subdomain,DC=company,DC=com' -Format DN

.Parameter Name
The SID (i.e. S-1-5-21-397955417-626881126-188441444-501) or the distinguished name (i.e. 'CN=user,OU=UserAccounts,DC=subdomain,DC=company,DC=com') to convert to a fully qualified UPN

.Parameter Type
The format of the value provided in the Name parameter (either 'SID' or 'DN')

.Example
PS >ConvertFrom-DirectoryServiceObjectName -Name 'S-1-5-21-397955417-626881126-188441444-501' -Format SID
user@subdomain.company.com

Returns the UPN name associated with a SID

.Example
PS >ConvertFrom-DirectoryServiceObjectName -Name 'CN=user,OU=UserAccounts,DC=subdomain,DC=company,DC=com' -Format DN
user@subdomain.company.com

Returns the UPN name associated with a distinguished name

#>
[CmdletBinding()]
param (
    [parameter(Mandatory=$true)]
    [string]$Name,

    [parameter(Mandatory=$true)]
    [ValidateSet("SID","DN")]
    [string]$Type
)

BEGIN {
Set-StrictMode -Version 1

$nativeMethodsSignature = @'
        /// <see cref="https://msdn.microsoft.com/en-us/library/ms675970(v=vs.85).aspx"/>
        [DllImport("ntdsapi.dll", CharSet = CharSet.Auto)]
        static public extern uint DsCrackNames(
           IntPtr hDS,
           int flags,
           int formatOffered,
           int formatDesired,
           uint cNames,
           string[] rpNames,
           out IntPtr ppResult
        );

        /// <see cref="https://msdn.microsoft.com/en-us/library/ms675978(v=vs.85).aspx"/>
        [DllImport("ntdsapi.dll", CharSet = CharSet.Auto)]
        static public extern void DsFreeNameResult(IntPtr pResult);

        /// <see cref="https://msdn.microsoft.com/en-us/library/ms676247(v=vs.85).aspx"/>
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
        public struct DS_NAME_RESULT_ITEM
        {
            public int status;
            public string pDomain;
            public string pName;
        }

        /// <see cref="https://msdn.microsoft.com/en-us/library/ms675931(v=vs.85).aspx"/>
        [DllImport("ntdsapi.dll", CharSet = CharSet.Auto)]
        static public extern uint DsBind(
            string DomainControllerName,
            string DnsDomainName,
            out IntPtr phDS
        );

        /// <see cref="https://msdn.microsoft.com/en-us/library/ms676053(v=vs.85).aspx"/>
        [DllImport("ntdsapi.dll", CharSet = CharSet.Auto)]
        static public extern uint DsUnBind(ref IntPtr phDS);

        /// <see cref="https://msdn.microsoft.com/en-us/library/ms676246(v=vs.85).aspx"/>
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
        public struct DS_NAME_RESULT
        {
            public uint cItems;
            public IntPtr rItems;
        }
        
        /// <see cref="https://msdn.microsoft.com/en-us/library/ms676245(v=vs.85).aspx"/>
        public enum DS_NAME_FORMAT
        {
            /**
             * Indicates that the fully qualified distinguished name is used. For example:
             * CN=someone,OU=Users,DC=Engineering,DC=Fabrikam,DC=Com
             **/
            DS_FQDN_1779_NAME = 1,

            /**
             * Indicates that it is using the user principal name (UPN). For example:
             * someone@engineering.fabrikam.com
             **/
            DS_USER_PRINCIPAL_NAME = 8,

            /**
             * Indicates a Security Identifier (SID) for the object.
             * This can be either the current SID or a SID from the object SID history.
             * The SID string can use either the standard string representation of a SID, or one of the string constants defined in Sddl.h.
             * For more information about converting a binary SID into a SID string, see SID Strings.
             * The following is an example of a SID string: 
             * S-1-5-21-397955417-626881126-188441444-501
             **/
            DS_SID_OR_SID_HISTORY_NAME = 11
        }

        /// <see cref="https://msdn.microsoft.com/en-us/library/ms676062(v=vs.85).aspx"/>
        public enum DS_NAME_FLAGS
        {
            /**
             * Enables cross forest trust referral.
             **/
            DS_NAME_FLAG_TRUST_REFERRAL = 8
        }

        /// <summary>
        /// The CrackName function converts a directory service object name to a fully qualified user principal name.
        /// </summary>
        /// <param name="name">The name of the directory service object to convert.</param>
        /// <param name="formatOffered">The format of the name of the directory service object.</param>
        /// <returns>The fully qualified UPN formatted name corresponding to the given directory service object.</returns>
        /// <exception cref="Win32Exception">Unexpected failures during p/invoke of DsBind or DsCrackNames.</exception>
        /// <exception cref="InvalidOperationException">Usually indicates the the parameter value of name did not match any directory service objects of the given format.</exception>
        /// <see cref="https://msdn.microsoft.com/en-us/library/ms675970(v=vs.85).aspx"/>
        /// <remarks>
        /// This is essentially a helper method that wraps the p/invoke calls to DsBind,DsUnbind, and DsCrackNames in a minimal and simple .NET function call.
        /// This is desirable as the caller of this method will be invoking it from Powershell.
        /// </remarks>
        public static string CrackName(string name, DS_NAME_FORMAT formatOffered)
        {
            IntPtr hDS;
            uint error = DsBind(null, null, out hDS);
            if (error != 0)
            {
                throw new Win32Exception(Convert.ToInt32(error), "DsBind returned an error");
            }
            try
            {
                string[] names = new string[] { name };
                IntPtr pResult;
                error = DsCrackNames(hDS,
                                     (int)DS_NAME_FLAGS.DS_NAME_FLAG_TRUST_REFERRAL,
                                     (int)formatOffered,
                                     (int)DS_NAME_FORMAT.DS_USER_PRINCIPAL_NAME,
                                     1, // cNames
                                     names, // rpNames
                                     out pResult);
                if (error != 0)
                {
                    throw new Win32Exception(Convert.ToInt32(error), "DsCrackNames returned an error");
                }
                try
                {
                    uint cItems = (uint)Marshal.ReadInt32(pResult);
                    if (cItems != 1)
                    {
                        throw new InvalidOperationException(String.Format("Expected one result from DsCrackNames, received {0}", cItems));
                    }
                    IntPtr curptr = Marshal.ReadIntPtr(pResult, Marshal.OffsetOf(typeof(DS_NAME_RESULT), "rItems").ToInt32());
                    DS_NAME_RESULT_ITEM result = (DS_NAME_RESULT_ITEM)Marshal.PtrToStructure(curptr, typeof(DS_NAME_RESULT_ITEM));
                    if (result.status != 0)
                    {
                        switch (result.status)
                        {
                            case 1: throw new InvalidOperationException("DS_NAME_ERROR_RESOLVING");
                            case 2: throw new InvalidOperationException("DS_NAME_ERROR_NOT_FOUND");
                            case 3: throw new InvalidOperationException("DS_NAME_ERROR_NOT_UNIQUE");
                            case 4: throw new InvalidOperationException("DS_NAME_ERROR_NO_MAPPING");
                            case 5: throw new InvalidOperationException("DS_NAME_ERROR_DOMAIN_ONLY");
                            case 6: throw new InvalidOperationException("DS_NAME_ERROR_NO_SYNTACTICAL_MAPPING");
                            case 7: throw new InvalidOperationException("DS_NAME_ERROR_TRUST_REFERRAL");
                            default: throw new InvalidOperationException(String.Format("DS_NAME_ERROR_UNKNOWN: {0}", result.status));
                        }
                    }
                    result.pName = String.IsNullOrWhiteSpace(result.pName) ? String.Empty : result.pName;
                    result.pDomain = String.IsNullOrWhiteSpace(result.pDomain) ? String.Empty : result.pDomain;
                    string[] nameParts = result.pName.Split('@');
                    if (nameParts.Length != 2 || String.IsNullOrWhiteSpace(nameParts[0]) || String.IsNullOrWhiteSpace(result.pDomain))
                    {
                        throw new InvalidOperationException(
                            String.Format(
                                "Expected result formatted like:  name:[alias@company.com], domain:[subdomain.company.com]. Was name:[{0}], domain:[{1}]",
                                result.pName,
                                result.pDomain
                            )
                        );
                    }
                    return String.Format("{0}@{1}", nameParts[0], result.pDomain);
                }
                finally
                {
                    DsFreeNameResult(pResult);
                }
            }
            finally
            {
                DsUnBind(ref hDS);
            }
        }
'@
    Add-Type -MemberDefinition $nativeMethodsSignature -Name NativeMethods -Namespace 'Microsoft.MBAM' -UsingNamespace 'System.ComponentModel'

    # Helper method that calls the C# CrackName method and pulls out the desired exception from the MethodInvocationException wrapper that Powershell unhelpfully added.
    Function CrackName {
        Param(
            [parameter(Mandatory=$true)]
            [String]$Name,
            [parameter(Mandatory=$true)]
            [Microsoft.MBAM.NativeMethods+DS_NAME_FORMAT]$Type
        )
        try {
            return [Microsoft.MBAM.NativeMethods]::CrackName($Name, $Type);
        } catch [System.Management.Automation.MethodInvocationException] {
            throw $_.Exception.InnerException
        }
    }# CrackName
    
    # Named definition of the ConvertFrom-DirectoryServiceObjectName cmdlet defined by this ps1 file.
    Function ConvertFrom-DirectoryServiceObjectName {
        [CmdletBinding()]
        Param (
            [parameter(Mandatory=$true)]
            [string]$Name,
            [parameter(Mandatory=$true)]
            [ValidateSet("SID","DN")]
            [string]$Type
        )
        BEGIN {}
        PROCESS {
            if ($Type -eq 'SID') {
                return CrackName -Name $Name -Type DS_SID_OR_SID_HISTORY_NAME
            }
            return CrackName -Name $Name -Type DS_FQDN_1779_NAME
        }
        END {}
    }
}# BEGIN
PROCESS{
    # If the caller is just trying to dot source this file then do not attempt to invoke the cmdlet.
    if (!('dotsource' -eq $Name)) {
        return ConvertFrom-DirectoryServiceObjectName @PSBoundParameters
    }
}
END {}

# SIG # Begin signature block
# MIIauQYJKoZIhvcNAQcCoIIaqjCCGqYCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUS7cJRKkef4+IyDS+ztdAzfJe
# VL2gghWCMIIEwzCCA6ugAwIBAgITMwAAAG9lLVhtBxFGKAAAAAAAbzANBgkqhkiG
# 9w0BAQUFADB3MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4G
# A1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEw
# HwYDVQQDExhNaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EwHhcNMTUwMzIwMTczMjAy
# WhcNMTYwNjIwMTczMjAyWjCBszELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hp
# bmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jw
# b3JhdGlvbjENMAsGA1UECxMETU9QUjEnMCUGA1UECxMebkNpcGhlciBEU0UgRVNO
# OkMwRjQtMzA4Ni1ERUY4MSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBT
# ZXJ2aWNlMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAz+ZtzcEqza6o
# XtiVTy0DQ0dzO7hC0tBXmt32UzZ31YhFJGrIq9Bm6YvFqg+e8oNGtirJ2DbG9KD/
# EW9m8F4UGbKxZ/jxXpSGqo4lr/g1E/2CL8c4XlPAdhzF03k7sGPrT5OaBfCiF3Hc
# xgyW0wAFLkxtWLN/tCwkcHuWaSxsingJbUmZjjo+ZpWPT394G2B7V8lR9EttUcM0
# t/g6CtYR38M6pR6gONzrrar4Q8SDmo2XNAM0BBrvrVQ2pNQaLP3DbvB45ynxuUTA
# cbQvxBCLDPc2Ynn9B1d96gV8TJ9OMD8nUDhmBrtdqD7FkNvfPHZWrZUgNFNy7WlZ
# bvBUH0DVOQIDAQABo4IBCTCCAQUwHQYDVR0OBBYEFPKmSSl4fFdwUmLP7ay3eyA0
# R9z9MB8GA1UdIwQYMBaAFCM0+NlSRnAK7UD7dvuzK7DDNbMPMFQGA1UdHwRNMEsw
# SaBHoEWGQ2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3Rz
# L01pY3Jvc29mdFRpbWVTdGFtcFBDQS5jcmwwWAYIKwYBBQUHAQEETDBKMEgGCCsG
# AQUFBzAChjxodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY3Jv
# c29mdFRpbWVTdGFtcFBDQS5jcnQwEwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZI
# hvcNAQEFBQADggEBAI2zTLbY7A2Hhhle5ADnl7jVz0wKPL33VdP08KCvVXKcI5e5
# girHFgrFJxNZ0NowK4hCulID5l7JJWgnJ41kp235t5pqqz6sQtAeJCbMVK/2kIFr
# Hq1Dnxt7EFdqMjYxokRoAZhaKxK0iTH2TAyuFTy3JCRdu/98U0yExA3NRnd+Kcqf
# skZigrQ0x/USaVytec0x7ulHjvj8U/PkApBRa876neOFv1mAWRDVZ6NMpvLkoLTY
# wTqhakimiM5w9qmc3vNTkz1wcQD/vut8/P8IYw9LUVmrFRmQdB7/u72qNZs9nvMQ
# FNV69h/W4nXzknQNrRbZEs+hm63SEuoAOyMVDM8wggTsMIID1KADAgECAhMzAAAB
# Cix5rtd5e6asAAEAAAEKMA0GCSqGSIb3DQEBBQUAMHkxCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xIzAhBgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNp
# Z25pbmcgUENBMB4XDTE1MDYwNDE3NDI0NVoXDTE2MDkwNDE3NDI0NVowgYMxCzAJ
# BgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25k
# MR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xDTALBgNVBAsTBE1PUFIx
# HjAcBgNVBAMTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJL8bza74QO5KNZG0aJhuqVG+2MWPi75R9LH7O3HmbEm
# UXW92swPBhQRpGwZnsBfTVSJ5E1Q2I3NoWGldxOaHKftDXT3p1Z56Cj3U9KxemPg
# 9ZSXt+zZR/hsPfMliLO8CsUEp458hUh2HGFGqhnEemKLwcI1qvtYb8VjC5NJMIEb
# e99/fE+0R21feByvtveWE1LvudFNOeVz3khOPBSqlw05zItR4VzRO/COZ+owYKlN
# Wp1DvdsjusAP10sQnZxN8FGihKrknKc91qPvChhIqPqxTqWYDku/8BTzAMiwSNZb
# /jjXiREtBbpDAk8iAJYlrX01boRoqyAYOCj+HKIQsaUCAwEAAaOCAWAwggFcMBMG
# A1UdJQQMMAoGCCsGAQUFBwMDMB0GA1UdDgQWBBSJ/gox6ibN5m3HkZG5lIyiGGE3
# NDBRBgNVHREESjBIpEYwRDENMAsGA1UECxMETU9QUjEzMDEGA1UEBRMqMzE1OTUr
# MDQwNzkzNTAtMTZmYS00YzYwLWI2YmYtOWQyYjFjZDA1OTg0MB8GA1UdIwQYMBaA
# FMsR6MrStBZYAck3LjMWFrlMmgofMFYGA1UdHwRPME0wS6BJoEeGRWh0dHA6Ly9j
# cmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY0NvZFNpZ1BDQV8w
# OC0zMS0yMDEwLmNybDBaBggrBgEFBQcBAQROMEwwSgYIKwYBBQUHMAKGPmh0dHA6
# Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljQ29kU2lnUENBXzA4LTMx
# LTIwMTAuY3J0MA0GCSqGSIb3DQEBBQUAA4IBAQCmqFOR3zsB/mFdBlrrZvAM2PfZ
# hNMAUQ4Q0aTRFyjnjDM4K9hDxgOLdeszkvSp4mf9AtulHU5DRV0bSePgTxbwfo/w
# iBHKgq2k+6apX/WXYMh7xL98m2ntH4LB8c2OeEti9dcNHNdTEtaWUu81vRmOoECT
# oQqlLRacwkZ0COvb9NilSTZUEhFVA7N7FvtH/vto/MBFXOI/Enkzou+Cxd5AGQfu
# FcUKm1kFQanQl56BngNb/ErjGi4FrFBHL4z6edgeIPgF+ylrGBT6cgS3C6eaZOwR
# XU9FSY0pGi370LYJU180lOAWxLnqczXoV+/h6xbDGMcGszvPYYTitkSJlKOGMIIF
# vDCCA6SgAwIBAgIKYTMmGgAAAAAAMTANBgkqhkiG9w0BAQUFADBfMRMwEQYKCZIm
# iZPyLGQBGRYDY29tMRkwFwYKCZImiZPyLGQBGRYJbWljcm9zb2Z0MS0wKwYDVQQD
# EyRNaWNyb3NvZnQgUm9vdCBDZXJ0aWZpY2F0ZSBBdXRob3JpdHkwHhcNMTAwODMx
# MjIxOTMyWhcNMjAwODMxMjIyOTMyWjB5MQswCQYDVQQGEwJVUzETMBEGA1UECBMK
# V2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0
# IENvcnBvcmF0aW9uMSMwIQYDVQQDExpNaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBD
# QTCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALJyWVwZMGS/HZpgICBC
# mXZTbD4b1m/My/Hqa/6XFhDg3zp0gxq3L6Ay7P/ewkJOI9VyANs1VwqJyq4gSfTw
# aKxNS42lvXlLcZtHB9r9Jd+ddYjPqnNEf9eB2/O98jakyVxF3K+tPeAoaJcap6Vy
# c1bxF5Tk/TWUcqDWdl8ed0WDhTgW0HNbBbpnUo2lsmkv2hkL/pJ0KeJ2L1TdFDBZ
# +NKNYv3LyV9GMVC5JxPkQDDPcikQKCLHN049oDI9kM2hOAaFXE5WgigqBTK3S9dP
# Y+fSLWLxRT3nrAgA9kahntFbjCZT6HqqSvJGzzc8OJ60d1ylF56NyxGPVjzBrAlf
# A9MCAwEAAaOCAV4wggFaMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFMsR6MrS
# tBZYAck3LjMWFrlMmgofMAsGA1UdDwQEAwIBhjASBgkrBgEEAYI3FQEEBQIDAQAB
# MCMGCSsGAQQBgjcVAgQWBBT90TFO0yaKleGYYDuoMW+mPLzYLTAZBgkrBgEEAYI3
# FAIEDB4KAFMAdQBiAEMAQTAfBgNVHSMEGDAWgBQOrIJgQFYnl+UlE/wq4QpTlVnk
# pDBQBgNVHR8ESTBHMEWgQ6BBhj9odHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtp
# L2NybC9wcm9kdWN0cy9taWNyb3NvZnRyb290Y2VydC5jcmwwVAYIKwYBBQUHAQEE
# SDBGMEQGCCsGAQUFBzAChjhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2Nl
# cnRzL01pY3Jvc29mdFJvb3RDZXJ0LmNydDANBgkqhkiG9w0BAQUFAAOCAgEAWTk+
# fyZGr+tvQLEytWrrDi9uqEn361917Uw7LddDrQv+y+ktMaMjzHxQmIAhXaw9L0y6
# oqhWnONwu7i0+Hm1SXL3PupBf8rhDBdpy6WcIC36C1DEVs0t40rSvHDnqA2iA6VW
# 4LiKS1fylUKc8fPv7uOGHzQ8uFaa8FMjhSqkghyT4pQHHfLiTviMocroE6WRTsgb
# 0o9ylSpxbZsa+BzwU9ZnzCL/XB3Nooy9J7J5Y1ZEolHN+emjWFbdmwJFRC9f9Nqu
# 1IIybvyklRPk62nnqaIsvsgrEA5ljpnb9aL6EiYJZTiU8XofSrvR4Vbo0HiWGFzJ
# NRZf3ZMdSY4tvq00RBzuEBUaAF3dNVshzpjHCe6FDoxPbQ4TTj18KUicctHzbMrB
# 7HCjV5JXfZSNoBtIA1r3z6NnCnSlNu0tLxfI5nI3EvRvsTxngvlSso0zFmUeDord
# EN5k9G/ORtTTF+l5xAS00/ss3x+KnqwK+xMnQK3k+eGpf0a7B2BHZWBATrBC7E7t
# s3Z52Ao0CW0cgDEf4g5U3eWh++VHEK1kmP9QFi58vwUheuKVQSdpw5OPlcmN2Jsh
# rg1cnPCiroZogwxqLbt2awAdlq3yFnv2FoMkuYjPaqhHMS+a3ONxPdcAfmJH0c6I
# ybgY+g5yjcGjPa8CQGr/aZuW4hCoELQ3UAjWwz0wggYHMIID76ADAgECAgphFmg0
# AAAAAAAcMA0GCSqGSIb3DQEBBQUAMF8xEzARBgoJkiaJk/IsZAEZFgNjb20xGTAX
# BgoJkiaJk/IsZAEZFgltaWNyb3NvZnQxLTArBgNVBAMTJE1pY3Jvc29mdCBSb290
# IENlcnRpZmljYXRlIEF1dGhvcml0eTAeFw0wNzA0MDMxMjUzMDlaFw0yMTA0MDMx
# MzAzMDlaMHcxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xITAf
# BgNVBAMTGE1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQTCCASIwDQYJKoZIhvcNAQEB
# BQADggEPADCCAQoCggEBAJ+hbLHf20iSKnxrLhnhveLjxZlRI1Ctzt0YTiQP7tGn
# 0UytdDAgEesH1VSVFUmUG0KSrphcMCbaAGvoe73siQcP9w4EmPCJzB/LMySHnfL0
# Zxws/HvniB3q506jocEjU8qN+kXPCdBer9CwQgSi+aZsk2fXKNxGU7CG0OUoRi4n
# rIZPVVIM5AMs+2qQkDBuh/NZMJ36ftaXs+ghl3740hPzCLdTbVK0RZCfSABKR2YR
# JylmqJfk0waBSqL5hKcRRxQJgp+E7VV4/gGaHVAIhQAQMEbtt94jRrvELVSfrx54
# QTF3zJvfO4OToWECtR0Nsfz3m7IBziJLVP/5BcPCIAsCAwEAAaOCAaswggGnMA8G
# A1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYEFCM0+NlSRnAK7UD7dvuzK7DDNbMPMAsG
# A1UdDwQEAwIBhjAQBgkrBgEEAYI3FQEEAwIBADCBmAYDVR0jBIGQMIGNgBQOrIJg
# QFYnl+UlE/wq4QpTlVnkpKFjpGEwXzETMBEGCgmSJomT8ixkARkWA2NvbTEZMBcG
# CgmSJomT8ixkARkWCW1pY3Jvc29mdDEtMCsGA1UEAxMkTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5ghB5rRahSqClrUxzWPQHEy5lMFAGA1UdHwRJ
# MEcwRaBDoEGGP2h0dHA6Ly9jcmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1
# Y3RzL21pY3Jvc29mdHJvb3RjZXJ0LmNybDBUBggrBgEFBQcBAQRIMEYwRAYIKwYB
# BQUHMAKGOGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljcm9z
# b2Z0Um9vdENlcnQuY3J0MBMGA1UdJQQMMAoGCCsGAQUFBwMIMA0GCSqGSIb3DQEB
# BQUAA4ICAQAQl4rDXANENt3ptK132855UU0BsS50cVttDBOrzr57j7gu1BKijG1i
# uFcCy04gE1CZ3XpA4le7r1iaHOEdAYasu3jyi9DsOwHu4r6PCgXIjUji8FMV3U+r
# kuTnjWrVgMHmlPIGL4UD6ZEqJCJw+/b85HiZLg33B+JwvBhOnY5rCnKVuKE5nGct
# xVEO6mJcPxaYiyA/4gcaMvnMMUp2MT0rcgvI6nA9/4UKE9/CCmGO8Ne4F+tOi3/F
# NSteo7/rvH0LQnvUU3Ih7jDKu3hlXFsBFwoUDtLaFJj1PLlmWLMtL+f5hYbMUVbo
# nXCUbKw5TNT2eb+qGHpiKe+imyk0BncaYsk9Hm0fgvALxyy7z0Oz5fnsfbXjpKh0
# NbhOxXEjEiZ2CzxSjHFaRkMUvLOzsE1nyJ9C/4B5IYCeFTBm6EISXhrIniIh0EPp
# K+m79EjMLNTYMoBMJipIJF9a6lbvpt6Znco6b72BJ3QGEe52Ib+bgsEnVLaxaj2J
# oXZhtG6hE6a/qkfwEm/9ijJssv7fUciMI8lmvZ0dhxJkAj0tr1mPuOQh5bWwymO0
# eFQF1EEuUKyUsKV4q7OglnUa2ZKHE3UiLzKoCG6gW4wlv6DvhMoh1useT8ma7kng
# 9wFlb4kLfchpyOZu6qeXzjEp/w7FW1zYTRuh2Povnj8uVRZryROj/TGCBKEwggSd
# AgEBMIGQMHkxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xIzAh
# BgNVBAMTGk1pY3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBAhMzAAABCix5rtd5e6as
# AAEAAAEKMAkGBSsOAwIaBQCggbowGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQw
# HAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwIwYJKoZIhvcNAQkEMRYEFGCH
# ttDiHHUS6k5bC+AzVF69hzz6MFoGCisGAQQBgjcCAQwxTDBKoCyAKgBNAGkAYwBy
# AG8AcwBvAGYAdAAgAEMAbwByAHAAbwByAGEAdABpAG8AbqEagBhodHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEBBQAEggEAE2WBsyTO6sEEXs4CtsdU
# xjzITo9tFmMMQ7xUDJFIeQAt2SZj8nsgIalv6bhfAu6vePHwWrmDwF2djw96i/A6
# Hg/onPieAPg53EWhfusGsbKg2T3/Xeez4qBFHP5jwwlpIdlg4FhTxQpikfOmwajR
# ZOIUPbSQBECpYXCHlTL18cOG4Z/bMSQEvdPzpwUz90DAzLrR8NffV95bD+IqPnOY
# tVK6IWQzlMLG39v8q0WPNSvQMbHBczJExb6RLs90L7F577PBqnGXKnXTUN4MqfWk
# 23tnhbEumZtGTSxpx8982UKj0nwDkgnpXbcs1wU6R7cQqkqZFDQ10AYI9cT83Fie
# 9KGCAigwggIkBgkqhkiG9w0BCQYxggIVMIICEQIBATCBjjB3MQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSEwHwYDVQQDExhNaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBQQ0ECEzMAAABvZS1YbQcRRigAAAAAAG8wCQYFKw4DAhoFAKBdMBgG
# CSqGSIb3DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTE1MDcxNjE5
# Mzg0NlowIwYJKoZIhvcNAQkEMRYEFLw79xoQv/tZJJM6D5XHKiMXAV96MA0GCSqG
# SIb3DQEBBQUABIIBAAB5mJYRR4CqQ4c7AOgeGgiGqle/C28/tz2YdZSFB4IJqbyV
# Cpz1BIFB1qBR8vrC9/3GyR7vA02c0O8H/bQbVUBWf0oRo6TvT3bkVtMVnlqxgzBJ
# dP8pk1bjLzUZeDVx+BdU25P8HWdB47r8LgFlfQhuF6XkHZzuQXWMzdJbrqamd9qf
# aDzyoLk3kY7bstHp/A8+8iD7t3zYl9D87OMzYlVRLfS6gWVP5jXRMDnvUkK6IzhE
# XugR0GMEVqpprjG6kmB/DrMgtizzu320E20ySUlDTCDxjZ5JSL2m887NtknTllCW
# QxnrFHMOmBvr+pQv+vwxzp5rVPGcL0tkcxtj1XA=
# SIG # End signature block
